/****************************************************************************
 *  Familia Lignum - Genealogical program                                   *
 *  Copyright (C) 2011-2012 Rafał Długołęcki <rafal@dlugolecki.net.pl>      *
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; version 2 of the License.                 *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License along *
 *  with this program; if not, write to the Free Software Foundation, Inc., *
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             *
 *                                                                          *
 ****************************************************************************/

#include <iostream>
#include <istream>
#include <sstream>
#include <sigc++/sigc++.h>
#include <glibmm/convert.h> //For Glib::ConvertError
#include <zipios++/fcollexceptions.h>
#include <zipios++/zipinputstream.h>
#include <zipios++/simplesmartptr.h>
#include <libxml++/libxml++.h>
#include <libxml/tree.h>

#include "gno.h"

#include "../individual.h"
#include "../family.h"
#include "../view.h"

using namespace std;
using namespace zipios;

namespace file {

Gno::Gno(File *file)
  : xmlpp::SaxParser(), m_ptrFile(file)
{
  set_substitute_entities(true);
  expect_characters = true;

  currentCollection = not_set;

  clog << "Gno/open" << endl;
  try {
    ZipInputStream is(m_ptrFile->details.path + "/" +
		      m_ptrFile->details.name + "." +
		      m_ptrFile->details.extension);
    parse_stream(is);
  }
  catch(IOException &e) {
    cerr << "IOException caught in main:" << endl;
    cerr << e.what() << endl;
  }
  catch(Exception &e) {
    cerr << "Some nasty error occured " << e.what() << endl;
  }
}

Gno::Gno(const String filepath)
  : xmlpp::SaxParser() {
  set_substitute_entities(true);
  expect_characters = true;

  currentCollection = not_set;

  clog << "Gno/open" << endl;
  try {
    ZipInputStream is(filepath);
    parse_stream(is);
  }
  catch(IOException &e) {
    cerr << "IOException caught in main:" << endl;
    cerr << e.what() << endl;
  }
  catch(Exception &e) {
    cerr << "Some nasty error occured " << e.what() << endl;
  }
}

Gno::~Gno() {
}

void Gno::on_start_element(const String& name,
                            const AttributeList& attributes) {
  //After each opening element, we could parse characters.
  //After closing element, no characters we will parse.
  expect_characters = true;

  if(name == "Individuals") {
    currentCollection = individual;
    individualState = individualStateUnknown;
  }
  else if(name == "Families") {
    currentCollection = family;
    familyState = familyStateUnknown;
  }

  if(name == "GenoMap") {
    FileView view;
    AttributeList::const_iterator it;
    for(it = attributes.begin(); it != attributes.end(); *it++) {
      view.parse(it->name, it->value); 
    }
    m_ptrFile->views.push_back(view);
  }

  switch(currentCollection) {
  case individual:
    on_individual_start_element(name, attributes);
    break;
  case family:
    on_family_start_element(name, attributes);
    break;
  }
}

void Gno::on_end_element(const Glib::ustring& n) {
  String name = Glib::locale_to_utf8(n);
  expect_characters = false;

  if(name == "Individual")  { individualState = individualStateUnknown; }
  else if(name == "Family") { familyState = familyStateUnknown; }
  else if((name == "Individuals") || (name == "Families")) {
    currentCollection = collectionTypeUnknown;
  }
}

void Gno::on_characters(const Glib::ustring& chars) {
  if(!expect_characters)
    return;

  switch(currentCollection) {
  case individual: {
    on_individual_characters(chars);
    break;
  }
  case family: {
    on_family_characters(chars);
    break;
  }
  }
}

void Gno::on_individual_start_element(const String& name,
				      const AttributeList& attributes) {
  if(name == "Individual") {
    Individual ind;
    ind.setId(parseIndividualIdentifier(attributes[0].value));
    m_ptrFile->individuals.push_back(ind);
  }

  if(name == "Name")         { individualState = individualName; }
  else if(name == "Display") { individualState = nameDisplay; }
  else if(name == "First")   { individualState = nameFirst; }
  else if(name == "Middle")  { individualState = nameMiddle; }
  else if(name == "Last")    { individualState = nameLast; }
  else if(name == "Pictures") { individualState = pictures; }
  else if(name == "Contacts") { individualState = contacts; }
  else if(name == "Position") { individualState = position;
    //If there are loaded genomaps, and no genomap is specified for
    //this individual. Add him to the first genomap.
    if(!m_ptrFile->views.empty()) {
      m_ptrFile->individuals.back().viewId() = m_ptrFile->views[0].name();
    }
      
    for(int i = 0; i < attributes.size(); i++) {
      if(attributes[i].name == "BoundaryRect") {
	m_ptrFile->individuals.back().boundary().parse(attributes[i].value, ',');
      }
      else if(attributes[i].name == "GenoMap") {
	m_ptrFile->individuals.back().viewId() = attributes[i].value;
      }
    }
  }
  else if(name == "Gender") { individualState = gender; }
  else if(name == "IsDead") { individualState = isDead; }
  else if(name == "Birth")  { individualState = birth; }
    else if(name == "Date") { individualState = birthDate; }
    else if(name == "Place") { individualState = birthPlace; }
      else if(name == "Baptism") { individualState = birthBaptism; }
      else if(name == "Date"
	      && individualState == birthBaptism) { individualState = birthBaptismDate; }
      else if(name == "Place") { individualState = birthBaptismPlace; }
      else if(name == "Officiator") { individualState = birthBaptismOfficiator; }
      else if(name == "Godfather") { individualState = birthBaptismGodfather; }
      else if(name == "Godmother") { individualState = birthBaptismGodmother; }
  else if(name == "Comment") { individualState = comment; }
  else if(name == "Educations") { individualState = educations; }
  else if(name == "Hyperlink") { individualState = hyperlink; }
  else if(name == "child_no") { individualState = child_no; }
}

void Gno::on_individual_characters(const Glib::ustring& chars) {
  switch(individualState) {
  case individualName:
  case nameDisplay:
    break;
  case nameFirst: {
    m_ptrFile->individuals.back().name().first += chars;
    break;
  }
  case nameMiddle: {
    m_ptrFile->individuals.back().name().middle += chars;
    break;
  }
  case nameLast: {
    m_ptrFile->individuals.back().name().last += chars;
    break;
  }
  case pictures:
  case contacts: {
    break;
  }
  case position: {
    m_ptrFile->individuals.back().position().parse(chars);
    break;
  }
  case gender: {
    m_ptrFile->individuals.back().setGender(parseIndividualGender(chars));
    break;
  }
  case isDead: {
    m_ptrFile->individuals.back().setDead(parseIndividualDeadState(chars));
    break;
  }
//  case birth:
  case birthDate:
  case birthPlace:
  case birthBaptism:
  case birthBaptismDate:
  case birthBaptismPlace:
//  case birthBaptismOfficiator:
  case birthBaptismGodfather:
  case birthBaptismGodmother:
  case comment:
  case educations:
//  case hyperlink:
  case child_no:
    break;
  }
}

void Gno::on_family_start_element(const String& name,
				  const AttributeList& attributes) {
  if(name == "Family") {
    Family fam;
    fam.setId(parseFamilyIdentifier(attributes[0].value));
    m_ptrFile->families.push_back(fam);
  }
}

void Gno::on_family_characters(const Glib::ustring& chars) {
  switch(familyState) {
//  case position:
//  case positionTop:
  case positionTopLeft:
  case positionTopRight:
//  case positionBottom:
  case positionBottomLeft:
  case positionBottomRight:
  case familyLine:
  case relation:
//  case children:
  case unions:
    break;
  }
}
int parseFamilyIdentifier(String identifier) {
  int id;
  stringstream(identifier.substr(3)) >> id;
  return id;
}

int parseIndividualIdentifier(String identifier) {
  int id;
  stringstream(identifier.substr(3)) >> id;
  return id;
}

gender_type parseIndividualGender(String str) {
  if(str == "M")      {
    return male;
  }
  else if(str == "F") {
    return female;
  }
  return unknownGender;
}

bool parseIndividualDeadState(String str) {
  if(str == "Y") {
    return true;
  }
  return false;
}

}
